from __future__ import absolute_import
from __future__ import print_function

import tangos as db
import tangos.core.timestep
from ..input_handlers import halo_stat_files as hsf
from .. import parallel_tasks
from ..log import logger
from . import GenericTangosTool


class PropertyImporter(GenericTangosTool):
    tool_name = 'import-properties'
    tool_description = 'Import a merger tree from the consistent-trees tool'

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')

        parser.add_argument('properties', action='store', nargs='+',
                            help="The names of the halo-finder pre-calculated properties to import")

        parser.add_argument('--backwards', action='store_true',
                            help='Process low-z timesteps first')

    def process_options(self, options):
        self.options = options

    def run_calculation_loop(self):
        base_sim = db.sim_query_from_name_list(self.options.sims)

        names = self.options.properties

        for x in base_sim:
            timesteps = db.core.get_default_session().query(tangos.core.timestep.TimeStep).filter_by(
                simulation_id=x.id, available=True).order_by(tangos.core.timestep.TimeStep.redshift.desc()).all()

            if self.options.backwards:
                timesteps = timesteps[::-1]

            for ts in parallel_tasks.distributed(timesteps):
                logger.info("Processing %s",ts)
                hsf.HaloStatFile(ts).add_halo_properties(*names)
                db.core.get_default_session().commit()



from __future__ import absolute_import
from __future__ import print_function

import tangos as db
import os
from ..input_handlers import consistent_trees as ct
from ..log import logger
from ..core import get_or_create_dictionary_item
from ..core.halo import PhantomHalo
from ..core.halo_data import HaloLink, HaloProperty
from .. import config
from . import GenericTangosTool
from six.moves import xrange
import re

class ConsistentTreesImporter(GenericTangosTool):
    tool_name = 'import-consistent-trees'
    tool_description = 'Import properties that were calculated by the halo finder'
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')

        parser.add_argument('--with-ids', action='store_true',
                            help="Store the consistent-trees ids as a property for each halo")


    def process_options(self, options):
        self.options = options


    @staticmethod
    def filename_to_snapnum(filename):
        match = re.match(".*(?:snapdir|snapshot)_([0-9]+)/?$", filename)
        if match:
            return int(match.group(1))
        else:
            raise ValueError("Unable to convert %s to snapshot number"%filename)

    def create_phantoms(self, timestep, n_phantoms):
        existing_phantoms = timestep.phantoms.all()
        existing_phantom_ids = []
        for p in existing_phantoms:
            existing_phantom_ids.append(p.finder_id)

        new_phantoms = []

        for i in xrange(1,n_phantoms+1):
            if -i not in existing_phantom_ids:
                new_ph = PhantomHalo(timestep, i, -i)
                new_phantoms.append(new_ph)

        session = db.core.get_default_session()

        session.add_all(new_phantoms)
        session.commit()
        logger.info("Add %d phantom halos to timestep %s", len(new_phantoms), timestep)
        logger.info("Total number of phantoms in tree %d; existing phantoms %d", n_phantoms, len(existing_phantoms))

    def create_timestep_halo_dictionary(self, ts):
        halos = ts.halos.all()
        phantoms = ts.phantoms.all()
        out = {}
        for h in halos:
            out[h.finder_id] = h
        for p in phantoms:
            out[p.finder_id] = p
        return out

    def create_links(self, ts, ts_next, link_dictionary):
        session = db.get_default_session()
        d_id = get_or_create_dictionary_item(session, "consistent_trees_link")
        objs_this = self.create_timestep_halo_dictionary(ts)
        objs_next = self.create_timestep_halo_dictionary(ts_next)
        links = []
        for this_id, (next_id, merger_ratio) in link_dictionary.items():
            this_obj = objs_this.get(this_id, None)
            next_obj = objs_next.get(next_id, None)
            if this_obj is not None and next_obj is not None:
                links.append(HaloLink(this_obj, next_obj, d_id, 1.0))
                links.append(HaloLink(next_obj, this_obj, d_id, merger_ratio))
        session.add_all(links)
        session.commit()
        logger.info("%d links created between %s and %s",len(links), ts, ts_next)

    def store_ids(self, ts, id_to_tree_id):
        session = db.get_default_session()
        objs = self.create_timestep_halo_dictionary(ts)
        dict_obj = get_or_create_dictionary_item(session, "consistent_trees_id")
        props = []
        for o in objs.values():
            tree_id = id_to_tree_id.get(o.finder_id, None)
            if tree_id is not None:
                props.append(HaloProperty(o, dict_obj, tree_id))
        session.add_all(props)
        session.commit()
        logger.info("%d consistent tree IDs added to step %s", len(props), ts)


    def run_calculation_loop(self):
        simulations = db.sim_query_from_name_list(self.options.sims)

        for simulation in simulations:
            logger.info("Processing %s",simulation)
            tree = ct.ConsistentTrees(os.path.join(config.base,simulation.basename))
            for ts in simulation.timesteps:
                snapnum = self.filename_to_snapnum(ts.extension)
                ts_next = ts.next

                if self.options.with_ids:
                    self.store_ids(ts, tree.get_finder_id_to_tree_id_for_snapshot(snapnum))

                if ts_next is not None:
                    n_phantoms = tree.get_num_phantoms_in_snapshot(snapnum+1)
                    self.create_phantoms(ts_next, n_phantoms)
                    self.create_links(ts, ts_next, tree.get_links_for_snapshot(snapnum))




from __future__ import absolute_import

import re
import numpy as np
import six
from ..log import logger


class BHShortenedLog(object):
    """Class to load a Changa BH .shortened.orbit file which contains the orbital and accretion histories of BHs"""
    _cache = {}

    @classmethod
    def get_existing_or_new(cls, filename):
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        obj = cls._cache.get(name, None)
        if obj is not None:
            return obj

        obj = cls(filename)
        cls._cache[name] = obj
        return obj

    def __init__(self, filename):
        import pynbody
        f = pynbody.load(filename)
        self.boxsize = float(f.properties['boxsize'].in_units('kpc', a=f.properties['a']))
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        ars = [[] for i in range(15)]
        for line in open(name + ".shortened.orbit"):
            line_split = line.split()
            for i in range(15):
                if i==0 or i==2:
                    col_type=int
                else:
                    col_type=float
                ars[i].append(col_type(line_split[i]))

        wrapped_ars = [pynbody.array.SimArray(x) for x in ars]
        for w in wrapped_ars:
            w.sim = f
        # bhid, time, step, mass, x, y, z, vx, vy, vz, pot, mdot, deltaM, E, dtEff, scalefac = wrapped_ars
        bhid, time, step, mass, x, y, z, vx, vy, vz, mdot, mdotmean, mdotsig, scalefac, dM = wrapped_ars
        bhid = np.array(bhid, dtype=int)
        logger.info("Loaded a BH log with %d entries", len(time))

        bhid[(bhid < 0)] = 2 * 2147483648 + bhid[(bhid < 0)]

        munits = f.infer_original_units("Msol")
        posunits = f.infer_original_units("kpc")
        velunits = f.infer_original_units("km s^-1")
        # potunits = velunits**2
        tunits = posunits / velunits
        # Eunits = munits*potunits
        # decorate with units

        x *= scalefac
        y *= scalefac
        z *= scalefac
        vx *= scalefac
        vy *= scalefac
        vz *= scalefac

        mass.units = munits
        x.units = y.units = z.units = posunits / pynbody.units.Unit('a')
        vx.units = vy.units = vz.units = velunits / pynbody.units.Unit('a')
        # pot.units = potunits
        time.units = tunits
        mdot.units = munits / tunits
        mdotsig.units = munits / tunits
        mdotmean.units = munits / tunits
        dM.units = munits
        # E.units = Eunits

        x.convert_units('kpc')
        y.convert_units('kpc')
        z.convert_units('kpc')
        vx.convert_units('km s^-1')
        vy.convert_units('km s^-1')
        vz.convert_units('km s^-1')
        mdot.convert_units('Msol yr^-1')
        mdotmean.convert_units('Msol yr^-1')
        mdotsig.convert_units('Msol yr^-1')
        mass.convert_units("Msol")
        time.convert_units("Gyr")
        dM.convert_units("Msol")
        # E.convert_units('erg')

        self.vars = {'bhid': bhid, 'step': step, 'x': x, 'y': y, 'z': z,
                     'vx': vx, 'vy': vy, 'vz': vz, 'mdot': mdot, 'mdotmean': mdotmean, 'mdotsig': mdotsig, 'mass': mass,
                     'time': time, 'dM': dM}

    def get_at_stepnum(self, stepnum):
        mask = self.vars['step'] == stepnum
        return dict((k, v[mask]) for k, v in six.iteritems(self.vars))

    def get_at_stepnum_for_id(self, stepnum, bhid):
        vars = self.get_at_stepnum(stepnum)
        try:
            index = np.where(vars['bhid']==bhid)[0][0]
        except IndexError:
            raise ValueError("BH %d not found in step %d"%(bhid,stepnum))
        vars_this = dict([(k,v[index]) for k, v in six.iteritems(vars)])
        return vars_this

    def get_last_entry_for_id(self, bhid):
        mask = self.vars['bhid'] == bhid
        if mask.sum()==0:
            raise ValueError("No entries for BH %d"%bhid)
        return dict((k, v[mask][-1]) for k, v in six.iteritems(self.vars))

    def determine_merger_ratio(self, bhid_eaten, bhid_survivor):
        eaten_entries = self.get_last_entry_for_id(bhid_eaten)
        eaten_mass = eaten_entries['mass']
        survivor_entries = self.get_at_stepnum_for_id(eaten_entries['step'], bhid_survivor)
        survivor_premerger_mass = survivor_entries['mass']
        return eaten_mass/survivor_premerger_mass


    def get_for_named_snapshot(self, filename):
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        stepnum = int(stepnum)
        return self.get_at_stepnum(stepnum)


from __future__ import absolute_import

import pynbody

class IsolatingSubSnapWrapper(pynbody.snapshot.SimSnap):
    """This class implements a SubSnap object which prevents access to its base.

    It is used to ensure consistent behaviour between analyses whether or not a full simulations snapshot is loaded
    See issue #42.
    """
    def __init__(self, subsnap):
        self._underlying = subsnap

    def __getattribute__(self, name):
        if name=='base' or name=='ancestor':
            return self
        return getattr(object.__getattribute__(self, "_underlying"), name)

    def __setattr__(self, key, value):
        setattr(object.__getattribute__(self, "_underlying"), key, value)

    def __delattr__(self, key):
        delattr(object.__getattribute__(self, "_underlying"), key)